--[[
    Istelien L (implementation of Istelien quiz engine in LUA with the aim of using it in LOTRO)
	Copyright (C) 2023-2024 Alejandro Meirinhos from Spain.

	This file is part of Istelien L.

	Istelien L is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Istelien L is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Istelien L.  If not, see <https://www.gnu.org/licenses/>.
--]]


-- Replace characters as per the saved configuration
function replaceCharacters(stringToReplace, replacementTable)

	for i in pairs(replacementTable) do
		stringToReplace = string.gsub(stringToReplace, i, replacementTable[i]) -- Simple lookup and replace through all elements of the list
	end

	return stringToReplace
end

-- Creates a string mask, which is the input string hiding all its characters except the first of each word.
function stringHint(inputString)

	wordList = stringSplit(inputString) -- Save a table with every word of the string

	local hint = ''
    for index in pairs(wordList) do -- Traverse the table joining back all the words while replacing characters
		hint = hint .. ' ' .. chatColours['Orange'] .. string.sub(wordList[index], 1, 1) .. chatColours['Closure']
		hint = hint .. string.gsub(string.sub(wordList[index], 2, #wordList[index]), '[^%p]', '%-') -- Replace everything except punctuation marks
	end
	hint = stringTrim(hint) -- Remove the unwanted spaces we introduced artificially

	return hint
end

-- Check if two strings match. By having a direct comparison before checking for each different difficulty setting we add those small checks but in return, if the match is done soon, we prevent the calling of the different functions, which are heavier.
function compareStrings(candidateString, targetString)
	local stringMatch
	stringMatch = false
	if candidateString == targetString then
			stringMatch = true
	else
		if (config['DIFFICULTY']['IgnoreCase'] == 1) then -- Transform strings to lower case
			candidateString = stringTrim(string.lower(candidateString))
			targetString = stringTrim(string.lower(targetString))
		end	
		if candidateString == targetString then
				stringMatch = true
		else
			if (config['DIFFICULTY']['IndulgentSpelling'] == 1) then -- Perform character substitutions (designed to ignore diacritics and accept many non-English characters)
				candidateString = stringTrim(replaceCharacters(candidateString, config['SPELLING']))
				targetString = stringTrim(replaceCharacters(targetString, config['SPELLING']))
			end	
			if candidateString == targetString then
					stringMatch = true
			else
				if (config['DIFFICULTY']['ApproximateAnswers'] == 1) then -- Implementation of fuzzy string match to allow non-exact answers to be accepted, thus permitting a certain degree of mistyping or grammar defects
					if stringProximity(candidateString, targetString) > (config['DIFFICULTY']['ProximityValue'] / 100) then
						stringMatch = true
					end
					-- Turbine.Shell.WriteLine('Answer Proximity: '.. stringProximity(candidateString, targetString)) -- DEBUG
				end
			end
		end
	end

	return stringMatch
end

-- Check if two strings match
function compareStringsPreviousVersion(candidateString, targetString)
	local stringMatch
	if (config['DIFFICULTY']['IgnoreCase'] == 1) then -- Transform strings to lower case
		candidateString = stringTrim(string.lower(candidateString))
		targetString = stringTrim(string.lower(targetString))
	end
	if (config['DIFFICULTY']['IndulgentSpelling'] == 1) then -- Perform character substitutions (designed to ignore diacritics and accept many non-English characters)
		candidateString = stringTrim(replaceCharacters(candidateString, config['SPELLING']))
		targetString = stringTrim(replaceCharacters(targetString, config['SPELLING']))
	end
	if (config['DIFFICULTY']['ApproximateAnswers'] == 1) then -- Implementation of fuzzy string match to allow non-exact answers to be accepted, thus permitting a certain degree of mistyping or grammar defects
		if stringProximity(candidateString, targetString) > (config['DIFFICULTY']['ProximityValue'] / 100) then
			stringMatch = true
		else
			stringMatch = false
		end
		-- Turbine.Shell.WriteLine('Answer Proximity: '.. stringProximity(candidateString, targetString)) -- DEBUG
	else
		if candidateString == targetString then
			stringMatch = true
		else
			stringMatch = false
		end
	end

	return stringMatch
end

--[[ 
Damerau-Levenshtein edit distance implementation from https://gist.github.com/badocelot/5327337
Based on pseudocode from Wikipedia: https://en.wikipedia.org/wiki/Damerau-Levenshtein_distance

Since I'm not skilled enough with such algorithm and the language, I've used a tool to convert it to LUA.
Its return values were always a fixed distance away from the target. Precisely because of my lack of skill,
I can only guess that this came from the conversion. I added a correction factor to bring it back to what 
would be the desired value.

--]]

function damerauLevenshteinDistance(a, b)
    -- "Infinity" -- greater than maximum possible edit distance
    -- Used to prevent transpositions for first characters
    local INF = #a + #b

    -- Matrix: (M + 2) x (N + 2)
    local matrix = {}
    for i = 1, #a + 2 do
        matrix[i] = {}
        for j = 1, #b + 2 do
            matrix[i][j] = INF
        end
    end

    for i = 0, #b do
        matrix[1][i+1] = i
    end
    for i = 0, #a do
        matrix[i+2][1] = i
    end

    -- Holds last row each element was encountered: `DA` in the Wikipedia pseudocode
    local last_row = {}
	
	-- From Istelien author: I added this because the values received were all a fixed distance away from what they should be.
	local corrFactor = -1

    -- Fill in costs
    for i = 1, #a do
        -- Current character in `a`
        local ch_a = string.sub(a, i, i)

        -- Column of last match on this row: `DB` in pseudocode
        local last_match_col = 0

        for j = 1, #b do
            -- Current character in `b`
            local ch_b = string.sub(b, j, j)

            -- Last row with matching character; `i1` in pseudocode
            local last_matching_row = last_row[ch_b] or 0

            -- Cost of substitution
            local cost = ch_a == ch_b and 0 or 1

            -- Compute substring distance
            matrix[i+2][j+2] = math.min(
                matrix[i+1][j+1] + cost, -- Substitution
                matrix[i+2][j+1] + 1,    -- Addition
                matrix[i+1][j+2] + 1,    -- Deletion

                -- Transposition
                matrix[last_matching_row+1][last_match_col+1]
                    + (i - last_matching_row - 1) + 1
                    + (j - last_match_col - 1))

            -- If there was a match, update last_match_col
            -- Doing this here lets me be rid of the `j1` variable from the original pseudocode
            if cost == 0 then
                last_match_col = j
            end
        end

        -- Update last row for current character
        last_row[ch_a] = i
    end

    -- Return last element
    return matrix[#a+2][#b+2] + corrFactor
end

-- Gets the likness between two strings by calculated the Damerau-Leventhstein Distance and turning it into a proximity coefficient
function stringProximity(a, b)
    return 1 - (damerauLevenshteinDistance(a, b) / (#a + #b))
end