--[[
    Istelien L (implementation of Istelien quiz engine in LUA with the aim of using it in LOTRO)
	Copyright (C) 2023-2024 Alejandro Meirinhos from Spain.

	This file is part of Istelien L.

	Istelien L is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Istelien L is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Istelien L.  If not, see <https://www.gnu.org/licenses/>.
--]]

-- Create a table from a string, spliting items per separator
function stringSplit(inputString, separator)

	if separator == nil then
			separator = '%s' -- Default separator is the space
	end

	local wordList = {}
	local wordCounter = 0
	for subString in string.gmatch(inputString, '([^' .. separator .. ']+)') do -- Extract a substring looking for the separator
		table.insert(wordList, subString)
		wordCounter = wordCounter + 1
	end

	return wordList, wordCounter
end

-- Function to remove leading and trailing spaces of a string
function stringTrim(s)
	local l = 1
	while string.sub(s, l, l) == ' ' do
		l = l + 1
	end
	local r = string.len(s)
	while string.sub(s, r, r) == ' ' do
		r = r - 1
	end
	return string.sub(s, l, r)
end

-- Add leading zeros (for date and time string pieces)
function addZeros(input, length)
	input = tostring(input)
	output = string.rep('0', length - #input) .. input
	return output
end

-- Function to round numbers to decimals
function round(number, decimals)
	if (decimals == nil) then
		decimals = 0
	end
    local power = 10 ^ decimals
    return math.floor(number * power) / power
end

function reverseNumber(number)
	str = tostring(math.floor(number))
	revNum = tonumber(string.reverse(str))
	return revNum
end

-- Shallow copy of one table into a new one
function copyTableShallow(input)
    local output = {}
    for k,v in pairs(input) do
        output[k] = v
    end
    return output
end

-- Deep copy of one table into a new one (from https://lua-users.org/wiki/CopyTable)
function copyTableDeep(input)
    local inputType = type(input)
    local output
    if inputType == 'table' then
        output = {}
        for inputKey, inputValue in next, input, nil do
            output[copyTableDeep(inputKey)] = copyTableDeep(inputValue)
        end
    else -- number, string, boolean, etc
        output = input
    end
    return output
end

-- Function to help enumerate a table by its keys (from Programming in LUA https://www.lua.org/pil/contents.html)
function pairsByKeys (t, f)
	local a = {}
	for n in pairs(t) do 
		table.insert(a, n)
	end
	table.sort(a, f)
	local i = 0      -- iterator variable
	local iter = function()   -- iterator function
			i = i + 1
			if (a[i] == nil) then 
				return nil
			else 
				return a[i], t[a[i]]
			end
		end
	return iter
end

-- Returns an array containing the sorted keys of a table
function sortedKeys(t)
	local sortedList = {}
	for key in pairsByKeys(t) do
      table.insert(sortedList, key)
    end
	return sortedList
end

-- Returns an array containing the input table sorted by its values (descending)
function sortedArrayFromTable(t)
	-- Convert t in Array
	local arrayFromTable = {}
	local counter = 0
	for i, v in pairs(t) do
		counter = counter + 1
		local pointsString = tostring(v)
		local maskedPointsString = '000000000000' .. pointsString
		local decimalSeparatorPos = 12 + #pointsString + 1
		local foundDot = string.find(maskedPointsString, '%.')
		if (foundDot ~= nil) then
			decimalSeparatorPos = foundDot
			maskedPointsString = maskedPointsString .. '00000000000000'
		else			
			maskedPointsString = maskedPointsString .. '.00000000000000'			
		end
		arrayFromTable[counter] = string.sub(maskedPointsString, decimalSeparatorPos - 12, decimalSeparatorPos + 12) .. ' ' .. i -- Concatenate value with index
	end
	table.sort(arrayFromTable, function(a, b) return a > b end) -- Sort
	-- Reconstruct the original keys and values in the new array
	for i,v in ipairs(arrayFromTable) do
		local vSplit = stringSplit(v)
		local oK = vSplit[2]
		local oV = tonumber(vSplit[1])
		arrayFromTable[i] = {[oK] = oV}
		-- Turbine.Shell.WriteLine('i: ' .. i .. ' - v: ' .. v .. ' - oK:' .. oK .. ' - oV:' .. oV) -- DEBUG
	end

	return arrayFromTable
end

-- Appends one array to another
function mergeArrays(array1, array2)
	local currentIndex = #array1
	for i,v in ipairs(array2) do
		currentIndex = currentIndex + 1
		array1[currentIndex] = v
	end
	return array1
end

-- Turn a string input into a number between certain values
function validateNumberString(inputString, minValue, maxValue, decimals)
	if (minValue == nil) then
		minValue = 0
	end
	if (maxValue == nil) then
		maxValue = 1000
	end	
	if (decimals == nil) then
		decimals = 0
	end
	local output = tonumber(inputString)
	if (output ~= nil) then
		output = round(output, decimals)
		if (output < minValue) then
			output = minValue
		end
		if (output > maxValue) then
			output = maxValue
		end	
	end	
	return output
end
