--[[
    Istelien L (implementation of Istelien quiz engine in LUA with the aim of using it in LOTRO)
	Copyright (C) 2023-2024 Alejandro Meirinhos from Spain.

	This file is part of Istelien L.

	Istelien L is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Istelien L is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Istelien L.  If not, see <https://www.gnu.org/licenses/>.
--]]

import 'Anglachel.Istelien.DefaultGui'
import 'Anglachel.Istelien.DefaultQuizLib'
import 'Anglachel.Istelien.DefaultSettings'
import 'Anglachel.Istelien.VindarPatch' -- Load/Save patch for European clients, written by Vindar: https://www.lotrointerface.com/downloads/info456-afixfortheparseerrorforeuropeanclients.html . And modified by Garan to solve the numeric formatting problem: https://forums-old.lotro.com/forums/showthread.php?428196-Writing-LoTRO-Lua-Plugins-for-Noobs&p=5784197#post5784197

-- Function to load and save data
function dataManagement(variable, operation, fileName)
	if (operation == 'Load') then
		variable = PatchDataLoad(Turbine.DataScope.Account, fileName)
	else
		if (operation == 'Save') then
			PatchDataSave(Turbine.DataScope.Account, fileName, variable)
		else
			return nil
		end
	end

	return variable
end

-- Initial set up of plugin settings
function initPluginData()
	local teams = {}
	
	gui = dataManagement(gui, 'Load', 'Istelien_gui')
	config = dataManagement(config, 'Load', 'Istelien_config')
	gameParticipants = dataManagement(gameParticipants, 'Load', 'Istelien_participants')

	-- Use default configuration if no data could be loaded
	if gui == nil then
		gui = defaultGui
		helpWindow(true)
		activateWindow(wHelpWindow, true)	
	end
	if config == nil then
		config = defaultSettings['CONFIG']
	end
	if (gameParticipants == nil) then
		gameParticipants = defaultSettings['PARTICIPANTS']
	end
	-- Quiz selection	
	quizLib = dataManagement(quizLib, 'Load', 'Istelien_quizLib')
	if (quizLib == nil) then -- If there's no library <yet> it's created from the default data, and settings are saved right away.
		quizLib = defaultQuizLib
		quizPackage = quizLib[1]
		config['GAME']['SelectedQuiz'] = {[1] = true}
		dataManagement(config, 'Save', 'Istelien_config')
		dataManagement(quizLib, 'Save', 'Istelien_quizLib')
	else
		quizPackage = loadQuiz(config['GAME']['SelectedQuiz']) -- Try to load the latest selection
		if (quizPackage == nil) then
			Turbine.Shell.WriteLine(sendToChat(replaceCharacters(translate('msgCouldNotLoadAnyPackage'), nonbasicCharactersInChat) .. '.', 'Append', '', true, 'Red'))
			quizPackage = defaultQuizLib[1]
		end
	end
	-- Savegames
	savegameLib = dataManagement(savegameLib, 'Load', 'Istelien_savegameLib')
	if (savegameLib == nil) then
		savegameLib = {}
	end
	
	gameState = 0
end

-- Prepare a list with the description (name and author) of each package contained in a library
function quizPackageList(library)
	local packageList = {}
	for i,v in ipairs(library) do
		local packageName = ''
		local packageAuthor = ''
		if (v['Info']['PackageName'] == nil) then
			packageName = translate('txtUnnamedPackage')
		else
			packageName = v['Info']['PackageName']
		end
		if (v['Info']['PackageCategory'] == nil) then			
			packageCategory = ''
		else
			packageCategory = ' - ' .. v['Info']['PackageCategory']
		end
		if (v['Info']['PackageAuthor'] == nil) then
			packageAuthor = translate('txtUnknownAuthor')
		else
			packageAuthor = v['Info']['PackageAuthor']
		end
		table.insert(packageList, {['Summary'] = '[' .. #v['Content'] .. '] ' .. packageName .. packageCategory .. ' ' .. translate('txtBy') .. ' ' .. packageAuthor, ['Questions'] = #v['Content']})
	end	
	
	return packageList
end

-- Return a quiz package (set of questions) from the library file according to a selection
function loadQuiz(indexList)
	local returnPackage = {} -- Base return value, to modify if there's some degree of success.
	local incompleteLoad = false
	-- Check which indexes are marked as selected. 
	local selectionCounter = 0
	local selectionList = {}
	
	for i,v in ipairs(indexList) do
		if (v == true) then
			table.insert(selectionList, i)
			selectionCounter = selectionCounter + 1
		end
	end	
	--[[ DEBUG
	for i,v in ipairs(selectionList) do
		Turbine.Shell.WriteLine('Selection i:' .. i .. ' - v:' .. tostring(v))
	end	
	for i,v in ipairs(quizLib) do
		Turbine.Shell.WriteLine('quizLib i:' .. i)
		Turbine.Shell.WriteLine('quizLib v:' .. tostring(v))
	end
	--]]
	
	if (selectionCounter == 1) then -- Simple selections simplify the process and would be more common so we check it first.
		returnPackage = quizLib[selectionList[1]]
	else
		if (selectionCounter > 1) then -- And leave merging as a separate case, which requires more complex processes
			returnPackage['Info'] = {['PackageName'] = '', ['PackageAuthor'] = ''}
			returnPackage['Content'] = {}
			local namesTable = {}
			local namesTableSize = 0
			local categoriesTable = {}
			local categoriesTableSize = 0
			local authorsTable = {}
			local authorsTableSize = 0
			for i,v in ipairs(selectionList) do 
				if (quizLib[v] ~= nil) then
					if (quizLib[v]['Content'] ~= nil) then
						-- Gather pack info. We put values as indexes, so that way we don't repeat anything.
						if (quizLib[v]['Info']['PackageName'] ~= nil) and (quizLib[v]['Info']['PackageName'] ~= '') then
							namesTable[quizLib[v]['Info']['PackageName']] = 1
							namesTableSize = namesTableSize + 1
						end
						if (quizLib[v]['Info']['PackageCategory'] ~= nil) and (quizLib[v]['Info']['PackageCategory'] ~= '') then
							categoriesTable[quizLib[v]['Info']['PackageCategory']] = 1
							categoriesTableSize = categoriesTableSize + 1
						end
						if (quizLib[v]['Info']['PackageAuthor'] ~= nil) and (quizLib[v]['Info']['PackageAuthor'] ~= '') then
							authorsTable[quizLib[v]['Info']['PackageAuthor']] = 1
							authorsTableSize = authorsTableSize + 1
						end
						-- Add pack content
						for j,w in ipairs(quizLib[v]['Content']) do
							table.insert(returnPackage['Content'], w)
						end
					end
				else
					incompleteLoad = true
				end
			end
			-- Check if we actually added questions, and if so, format the combined package's info
			if (#returnPackage['Content'] > 0) then
				if (namesTableSize > 0) then -- We'd like to preserve the package names if they exist
					local namesList = sortedKeys(namesTable)
					returnPackage['Info']['PackageName'] = ''
					for i,v in ipairs(namesList) do
						returnPackage['Info']['PackageName'] = returnPackage['Info']['PackageName'] .. ', ' .. v 
					end
					returnPackage['Info']['PackageName'] = string.sub(returnPackage['Info']['PackageName'], 3, #returnPackage['Info']['PackageName']) -- Remove initial comma and space
					if (#returnPackage['Info']['PackageName'] > 60) then -- Too long
						returnPackage['Info']['PackageName'] = 'Compilation'
					end
				else
					returnPackage['Info']['PackageName'] = 'Compilation'
				end
				if (categoriesTableSize > 0) then -- We'd like to preserve the package names if they exist
					local categoriesList = sortedKeys(categoriesTable)
					returnPackage['Info']['PackageCategory'] = ''
					for i,v in ipairs(categoriesList) do
						returnPackage['Info']['PackageCategory'] = returnPackage['Info']['PackageCategory'] .. ', ' .. v 
					end
					returnPackage['Info']['PackageCategory'] = string.sub(returnPackage['Info']['PackageCategory'], 3, #returnPackage['Info']['PackageCategory']) -- Remove initial comma and space
					if (#returnPackage['Info']['PackageCategory'] > 60) then -- Too long
						returnPackage['Info']['PackageCategory'] = 'Assorted'
					end
				else
					returnPackage['Info']['PackageCategory'] = 'Assorted'
				end
				if (authorsTableSize > 0) then -- If there are authors we want to continue giving them credit
					local authorList = sortedKeys(authorsTable)
					returnPackage['Info']['PackageAuthor'] = ''
					for i,v in ipairs(authorList) do
						returnPackage['Info']['PackageAuthor'] = returnPackage['Info']['PackageAuthor'] .. ', ' .. v 
					end
					returnPackage['Info']['PackageAuthor'] = string.sub(returnPackage['Info']['PackageAuthor'], 3, #returnPackage['Info']['PackageAuthor']) -- Remove initial comma and space
					if (#returnPackage['Info']['PackageAuthor'] > 60) then -- Too long
						returnPackage['Info']['PackageAuthor'] = 'Multiauthored'
					end
				else
					returnPackage['Info']['PackageAuthor'] = 'Istelien'
				end
			else -- In case I'm missing something, I'll ensure we return nil because what's important is to have questions
				returnPackage = nil
			end
		else 
			returnPackage = nil
		end
	end	
	if (returnPackage ~= nil) and (incompleteLoad == true) then
		Turbine.Shell.WriteLine(sendToChat(replaceCharacters(translate('msgCouldNotLoadSomePackage'), nonbasicCharactersInChat) .. '.', 'Append', '', true, 'Red'))
	end
	return returnPackage
end

function createTeam(teamName)
	gameParticipants['Teams'][teamName] = {['Points'] = 0,['Members'] = {}}
end

function deleteTeam(teamName)
	if (gameParticipants['Teams'][teamName] ~= nil) then
		if (gameParticipants['Teams'][teamName]['Members'] ~= nil) then
			for i,v in pairs(gameParticipants['Teams'][teamName]['Members']) do
				removePlayerFromTeam(i)
			end
		end
		gameParticipants['Teams'][teamName] = nil
	end
end

-- Deletes any player not in a team and any team without members, and register players already declared in teams
function resetParticipants()
	-- Go through players
	for k,v in pairs(gameParticipants['Players']) do
		if (k['Team'] == nil) then -- Delete player if it doesn't have team
			gameParticipants['Players'][k] = nil
		end
	end
	-- Go through teams
	for k,v in pairs(gameParticipants['Teams']) do
		local memberCounter = 0
		for i,w in pairs(gameParticipants['Teams'][k]['Members']) do
			if (i ~= nil) then
				memberCounter = memberCounter + 1
				if (gameParticipants['Players'][w] == nil) then
					updatePlayerData(i, k) -- Register player if it doesn't exist
				end
			end 
		end
		if (memberCounter == 0) then
			gameParticipants['Teams'][k] = nil -- Delete team if it doesn't have members
		end
	end
end

-- Adds a player to a team
function addPlayerToTeam(playerName, teamName)
	if (gameParticipants['Players'][playerName] == nil) then -- Player was not registered yet
		updatePlayerData(playerName)
	end
	if (config['GAME']['TeamsLock'] == 1) then
		return sendToChat(replaceCharacters(translate('msgTeamChangesNotAllowed'), nonbasicCharactersInChat) .. '.', 'Tell', playerName, true)
	else
		if (teamName == nil) then
			return sendToChat(replaceCharacters(translate('msgCouldNotJoinTeam'), nonbasicCharactersInChat) .. ': ' .. chatColours['Orange'] .. '!jointeam <Team name>' .. chatColours['Closure'] .. '.', 'Tell', playerName, true)
		else
			if (gameParticipants['Teams'][teamName] ~= nil) then -- Add player
				gameParticipants['Teams'][teamName]['Members'][playerName] = 1
			else -- Create team from scratch
				createTeam(teamName)
				gameParticipants['Teams'][teamName]['Members'] = {[playerName] = 1}
			end
			gameParticipants['Players'][playerName]['Team'] = teamName
			return sendToChat(replaceCharacters(translate('msgJoinedTeam'), nonbasicCharactersInChat) .. ' ' .. chatColours['Orange'] .. teamName .. chatColours['Closure'] .. '.', 'Tell', playerName, true)
		end
	end
end

-- Removes a player from team (if any)
function removePlayerFromTeam(playerName)
	if (gameParticipants['Players'][playerName] == nil) then -- Player was not registered yet
		updatePlayerData(playerName)
	end
	if (gameParticipants['Players'][playerName]['Team'] ~= nil) then
		local teamToLeave = gameParticipants['Players'][playerName]['Team']
		gameParticipants['Teams'][teamToLeave]['Members'][playerName] = nil
		gameParticipants['Players'][playerName]['Team'] = nil -- Delete ['Team'] key
		return sendToChat(replaceCharacters(translate('msgLeftTeam'), nonbasicCharactersInChat) .. '.', 'Tell', playerName, true)		
	else
		return sendToChat(replaceCharacters(translate('msgAlreadyWithoutTeam'), nonbasicCharactersInChat) .. '.', 'Tell', playerName, true)		
	end
end

-- Manage participants and teams
function updatePlayerData(playerName, teamName)

	if (gameParticipants['Players'][playerName]) == nil then -- Register player if it doesn't exist yet
		-- Turbine.Shell.WriteLine('split 1: '.. playerName) -- DEBUG
		gameParticipants['Players'][playerName] = {["Points"] = 0}
	end
	if (teamName ~= nil) then -- The argument has been provided
		-- Turbine.Shell.WriteLine('split 1: '.. playerName .. ' - split 2: ' .. teamName) -- DEBUG
		removePlayerFromTeam(playerName)
		if (teamName ~= '') then -- Add player to a team
			addPlayerToTeam(playerName, teamName)
		end
	end
end

-- Wipe a player
function deletePlayer(playerName)
	removePlayerFromTeam(playerName)
	gameParticipants['Players'][playerName] = nil
end

-- Retrieves the list of participants ordered alphabetically
function listParticipants(participantsType)
	return sortedKeys(gameParticipants[participantsType])
end

-- Enumerate team members
function listTeamMembers(playerName, teamName)
	local teamsList = {}
	if (teamName == nil) then -- If no team specified, use the player's team.
		if (playerName ~= nil) and (playerName ~= '') then
			teamName = gameParticipants['Players'][playerName]['Team']
		end
	end
	if (teamName~= nil) then
		if (gameParticipants['Teams'][teamName] ~= nil) then
			teamsList = sortedKeys(gameParticipants['Teams'][teamName]['Members'])
		end
	end
	
	return teamsList
end

-- Transfer a saved game from a saved games library into the active global variables and check for sanity pre and post operation
function loadGame(library, index)
	loadSuccess = false -- Base value
	
	-- Retrieve and validate information from the library
	local tempConfig = savegameLib[index]['CONFIG']
	local tempGameParticipants = savegameLib[index]['PARTICIPANTS']
	local tempGameLog = savegameLib[index]['GAMELOG']
	if (tempConfig ~= nil)  and (tempGameParticipants ~= nil) and (tempGameLog ~= nil) then -- Tables were well retrieved
		if (tempConfig['GAME'] ~= nil) then
			if (tempGameLog['QuestionCounter'] ~= nil) and (tempConfig['GAME']['GameMode'] ~= nil) and (tempConfig['GAME']['SelectedQuiz'] ~= nil) then -- Specific important values are well retrieved
				if (tempConfig['GAME']['GameMode'] == 2) then								
					if (tempGameLog['GameCountdown'] ~= nil) then -- One last value if the game is timed.
						loadSuccess = true
					end
				else
					loadSuccess = true
				end
			end
		end
	end
	
	-- Transfer to the active global variables and validate again. Yes, I'm being overly cautious, but all for data safey.
	if (loadSuccess == true) then
		config = copyTableDeep(tempConfig)
		gameParticipants = copyTableDeep(tempGameParticipants)
		gameLog = copyTableDeep(tempGameLog)
		questionCounter = gameLog['QuestionCounter']
		pointGiftCounter = 0	
		if (gameLog[0] ~= nil) then
			if (gameLog[0]['PointGiftCounter'] ~= nil) then
				pointGiftCounter = gameLog[0]['PointGiftCounter']
			end
		else	
			gameLog[0] = {}
		end
		if (config ~= nil)  and (gameParticipants ~= nil) and (gameLog ~= nil) then -- 
			if (config['GAME']['GameMode'] == 2) then
				gameCountdownSave = gameLog['GameCountdown']
				if (gameCountdownSave == nil) then
					loadSuccess = false
				end
			end
		else
			loadSuccess = false
		end
		-- Load Quiz Package
		quizPackage = loadQuiz(config['GAME']['SelectedQuiz'])
		if (quizPackage == nil) then
			loadSuccess = false
		end		
	end
	
	return loadSuccess
end

-- Transfer the global game variables to a saved game table to include it into the saved games file later
function saveGame(index, name)
	saveSuccess = false -- Base value
	local savegameLibBackup = copyTableDeep(savegameLib)
	
	local savedGame = {}
	if (config['GAME']['GameMode'] == 2) then
		gameLog['GameCountdown'] = gameCountdownSave
	end
	local dateTable = Turbine.Engine.GetDate()
	savedGame['SaveName'] = name .. ' - ' .. addZeros(dateTable['Year'], 4) .. '-' .. addZeros(dateTable['Month'], 2) .. '-' .. addZeros(dateTable['Day'], 2) .. 'T' .. addZeros(dateTable['Hour'], 2) .. ':' .. addZeros(dateTable['Minute'], 2) .. ':' .. addZeros(dateTable['Second'], 2) -- Attach date and time of save.
	savedGame['CONFIG'] = copyTableDeep(config)
	savedGame['PARTICIPANTS'] = copyTableDeep(gameParticipants)
	savedGame['GAMELOG'] = copyTableDeep(gameLog)
	if (savedGame['SaveName'] ~= nil) and (savedGame['SaveName'] ~= nil) and (savedGame['SaveName'] ~= nil) and (savedGame['SaveName'] ~= nil) then
		savegameLib[index] = savedGame
		if (savegameLib[index] ~= nil) then
			saveSuccess = true	
		else
			savegameLib = copyTableDeep(savegameLibBackup)
		end
	end
	
	return saveSuccess
end

--[[ Wannabe functions for merging quiz libraries, to save the user the minimum hassle of opening and writing files. Not used or tested yet because such feature has not been properly thought and designed.
-- Merges two quizzes
function mergeQuizContents(quiz1, quiz2, newName, newAuthor)
	if (quiz1 ~= nil and quiz2 ~= nil) then
		if (newName == nil) then
			newName = 'Merged quiz'
		end	
		if (newAuthor == nil) then
			newAuthor = 'Istelien'
		end
		local newContent = mergeArrays(quiz1['Content'], quiz2['Content'])
		
		-- Tweak names and authors
		if (quiz1['Info']['PackageName'] ~= '') then
			newName = newName .. ': ' .. quiz1['Info']['PackageName']
		end
		if (quiz1['Info']['PackageName'] ~= '' and quiz2['Info']['PackageName'] ~= '') then
			newName = newName .. ' & '
		end
		if (quiz2['Info']['PackageName'] ~= '') then
			newName = newName .. quiz2['Info']['PackageName']
		end
		if (quiz1['Info']['PackageAuthor'] ~= '') then
			newAuthor = newAuthor .. ', ' .. quiz1['Info']['PackageAuthor']
		end
		if (quiz2['Info']['PackageAuthor'] ~= '') then
			newAuthor = newAuthor .. ', ' .. quiz2['Info']['PackageAuthor']
		end
		
		newContent['Info'] = {['PackageName'] = newName, ['PackageAuthor'] = newAuthor}
				
		return newContent
	else
		return nil
	end
end

-- Merges two quiz packages (what is inside the files)
function mergeQuizPackages(pack1, pack2)
	if (pack1 ~= nil and pack2 ~= nil) then
		
		return mergeArrays(pack1, pack2)
	else
		return nil
	end
end

-- Merges two quiz files
function mergeQuizFiles(file1, file2, newFile)
	if (newFile ~= nil) then
		newFile = 'Istelien_QuizPackage_Merged'
	end
	local pack1 = dataManagement(pack1, 'Load', file1)
	local pack2 = dataManagement(pack2, 'Load', file2)
	if (pack1 ~= nil and pack2 ~= nil) then
		local newPack = mergeArrays(pack1, pack2)
	end
	
	return dataManagement(newPack, 'Save', newFile)
end
--]]