--[[
    Istelien L (implementation of Istelien quiz engine in LUA with the aim of using it in LOTRO)
	Copyright (C) 2023-2024 Alejandro Meirinhos from Spain.

	This file is part of Istelien L.

	Istelien L is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Istelien L is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Istelien L.  If not, see <https://www.gnu.org/licenses/>.
--]]

-- Basic timer class from LotroInterface (Garan)
Timer = class( Turbine.UI.Control );
function Timer:Constructor()
	Turbine.UI.Control.Constructor( self );
	self.EndTime=Turbine.Engine.GetGameTime();
	self.Repeat=false;

	self.SetTime=function(sender, numSeconds, setRepeat)
		numSeconds=tonumber(numSeconds);
		if numSeconds==nil or numSeconds<=0 then
			numSeconds=0;
		end
		self.EndTime=Turbine.Engine.GetGameTime()+numSeconds;
		self.Repeat=false; -- default
		self.NumSeconds=numSeconds;
		if setRepeat~=nil and setRepeat~=false then
			-- any non-false value will trigger a repeat
			self.Repeat=true;
		end
		self:SetWantsUpdates(true);
	end

	self.Update=function()
		if self.EndTime~=nil and Turbine.Engine.GetGameTime()>=self.EndTime then
			-- turn off timer to avoid firing again while processing
			self:SetWantsUpdates(false);
			-- fire whatever event you are trying to trigger
			if self.TimeReached~=nil then
				if type(self.TimeReached)=="function" then
					self.TimeReached();
				elseif type(self.TimeReached)=="table"  then
					for k,v in pairs(self.TimeReached) do
						if type(v)=="function" then
							v();
						end
					end
				end
			end

			if self.Repeat then
				self.EndTime=Turbine.Engine.GetGameTime()+self.NumSeconds;
				self:SetWantsUpdates(true);
			end
		end
	end
end

-- Countdown for when the game is timed.
gameTimerEvent = function()
		if (gameCountdown ~= nil) then
			-- Substract one second
			gameCountdown = gameCountdown - 1
			-- Turbine.Shell.WriteLine('Game countdown: ' .. tostring(gameCountdown)) -- DEBUG
			if (gameCountdown == 0) then -- Game time is over
				finishGame = true -- This indicates that the game should stop
			end

			return
		end
	end

-- Countdown for the active question
questionTimerEvent = function()
		if (questionCountdown ~= nil) then
			-- Substract one second
			questionCountdown = questionCountdown - 1
			-- Turbine.Shell.WriteLine('Question countdown: ' .. tostring(questionCountdown) .. ' : Question time = ' .. tostring(config['GAME']['QuestionTime']) .. ' - Question Hint = ' .. tostring(config['GAME']['QuestionHint']) .. ' - Result = ' .. tostring(config['GAME']['QuestionTime'] - config['GAME']['QuestionHint'])) -- DEBUG
			if (questionCountdown <= 0) then -- Question time is over

				-- Stop this timer
				questionTimer, questionTimerEvent, questionCountdown = closeTimer(questionTimer, questionTimerEvent, questionCountdown)

				endQuestion()
			else
				if (questionCountdown == (config['GAME']['QuestionTime'] - config['GAME']['QuestionHint'])) then -- We've reached the time to throw a hint
					answerHint = stringHint(questionAnswer) -- Obtain hint mask

					-- Throw hint to the chat
					-- Turbine.Shell.WriteLine(answerHint) -- DEBUG
					wMainWindow.qsProvideHint:SetEnabled(true)
					wMainWindow.qsProvideHint:SetVisible(true)
					wMainWindow.btnProvideHint:SetEnabled(true)
					wMainWindow.btnProvideHint:SetVisible(true)
					remapShortcut(wMainWindow.qsProvideHint, sendToChat(chatColours['Orange'] .. replaceCharacters(translate('msgHint'), nonbasicCharactersInChat) .. chatColours['Closure'] .. ': ' .. answerHint, config['GAME']['GameChannel'], ''))
				end
			end
		end

		return
	end

-- Countdown for additional answer entries
extraTimerEvent = function()
		if (extraCountdown ~= nil) then -- There's already a timer for this phase
			-- Substract one second
			extraCountdown = extraCountdown - 1
			-- Turbine.Shell.WriteLine('Extra countdown: ' .. tostring(extraCountdown)) -- DEBUG
			if (extraCountdown <= 0) then -- Time's up
				-- Turbine.Shell.WriteLine('Closing Extra in its own')
				-- Stop this timer
				extraTimer, extraTimerEvent, extraCountdown = closeTimer(extraTimer, extraTimerEvent, extraCountdown)

				endQuestion()
			end
		end
	end

-- Ends a timer if it's ongoing
function closeTimer(customTimer, customtimerEvent, countdownVar)
	-- Turbine.Shell.WriteLine('Closing timer (function start)') -- DEBUG
	if customTimer ~= nil then
		-- Turbine.Shell.WriteLine('Closing timer (inside if)') -- DEBUG
		removeCallback(customTimer, 'TimeReached', customtimerEvent)
		customTimer.Repeat = false
		customTimer = nil
		countdownVar = nil
	end
	return customTimer, customtimerEvent, countdownVar
end

function indexPackage()
	packageName = quizPackage['Info']['PackageName']
	packageCategory = quizPackage['Info']['PackageCategory']
	packageAuthor = quizPackage['Info']['PackageAuthor']
	maxQuestionIndex = #quizPackage['Content']
	-- Create a seed for the random selection of questions
	gDate = Turbine.Engine.GetDate()
	rSeed = reverseNumber(Turbine.Engine.GetGameTime() / Turbine.Engine.GetLocalTime() * tonumber(tostring(gDate['Year']) .. tostring(gDate['DayOfYear'])) / gDate['DayOfWeek'] / gDate['Month'] * 100000) + tonumber((gDate['Minute'] .. tostring(tonumber(gDate['Hour']) + tonumber(gDate['Second'])))) * (tonumber(gDate['Minute']) + tonumber(gDate['Second']))
	math.randomseed(rSeed)
	-- Turbine.Shell.WriteLine('Seed:'.. rSeed) -- DEBUG
end

function startGame()
	resetParticipants()
	resetScores() -- Reset scores from participant data
	resetStartGameMessage()
	-- Create applicable timers
	if (config['GAME']['GameMode'] == 2) then -- Timed game
		gameTimer = Timer()
		gameCountdown = config['GAME']['GameTarget'] * 60
		addCallback(gameTimer, 'TimeReached', gameTimerEvent)
		gameTimer:SetTime(1,true)
	end
	Turbine.Shell.WriteLine(istelienTextLogo .. ' '.. replaceCharacters(replaceCharacters(translate('msgGameStart'), nonbasicCharactersForGuiPrev), nonbasicCharactersForGui) .. '.')
	gameLog = {}
	gameLog[0] = {} -- Space for point gifts
	questionCounter = 0
	pointGiftCounter = 0
	
	indexPackage()
	
	-- GUI options lock
	lockMainGui(true)
	-- GUI game controls change
	extendMainGui(true)
	if (config['GAME']['TeamsLock'] == 1) then
		wMainWindow.btnManageTeams:SetEnabled(false)
	else
		wMainWindow.btnManageTeams:SetEnabled(true)
	end
	-- wMainWindow.btnStartGame:SetText(translate('txtPauseGame'))
	wMainWindow.qsStartGame:SetEnabled(false)
	wMainWindow.qsStartGame:SetVisible(false)
	wMainWindow.btnStartGame:SetEnabled(false)
	wMainWindow.btnStartGame:SetVisible(false)
	wMainWindow.qsPauseGame:SetEnabled(true)
	wMainWindow.qsPauseGame:SetVisible(true)
	wMainWindow.btnPauseGame:SetEnabled(true)
	wMainWindow.btnPauseGame:SetVisible(true)
	wMainWindow.btnEndGame:SetEnabled(true)
	wMainWindow.btnLoadSave:SetEnabled(false)
	wMainWindow.btnLoadSave:SetText(translate('txtSaveGame'))
	wMainWindow.btnAnnounceScores:SetEnabled(true)
	wMainWindow.btnAnnounceScores:SetVisible(true)
	wMainWindow.btnPointGift:SetEnabled(true)
	wMainWindow.btnPointGift:SetVisible(true)
	if (config['GAME']['QuestionMode'] == 1) then
		wMainWindow.btnMasterScorer:SetVisible(false)
	else
		wMainWindow.btnMasterScorer:SetVisible(true)
	end
	remapShortcut(wMainWindow.qsEndQuestion, '/dance')
	wMainWindow.qsEndQuestion:SetVisible(true)
	wMainWindow.btnEndQuestion:SetVisible(true)
	wMainWindow.btnEndQuestion:SetEnabled(false)
	wMainWindow.qsPauseGame:SetEnabled(false)
	wMainWindow.btnPauseGame:SetEnabled(false)
	wMainWindow.qsAskQuestion:SetEnabled(true)
	wMainWindow.qsAskQuestion:SetVisible(true)
	wMainWindow.btnAskQuestion:SetEnabled(true)
	wMainWindow.btnAskQuestion:SetVisible(true)
	
	-- Start rolling
	gameState = 1 -- 0 = Game is not running. 1 = Running. 2 = Paused.
	questionState = 0 -- 0 = No questions running. 1 = Running and allowing answers. 2 = Not allowing answers, interval to wrap question. 
	finishGame = false
	preQuestion() -- Launch first question
end

function pauseGame()
	Turbine.Shell.WriteLine(istelienTextLogo .. ' ' .. replaceCharacters(replaceCharacters(translate('msgGamePause'), nonbasicCharactersForGuiPrev), nonbasicCharactersForGui) .. '.')
	gameState = 2
	wMainWindow.qsPauseGame:SetEnabled(false)
	wMainWindow.btnPauseGame:SetEnabled(false)
	wMainWindow.btnLoadSave:SetEnabled(true)
	if (questionState == 0) then --We'ce clicked it after we wrapped a question, but before starting the next
		questionCounter = questionCounter - 1 -- We have to rewind because in this specific condition, the counter has already advanced as preQuestion() was already run
		preQuestion()
	end
end

function pausedGameFromSave()
	Turbine.Shell.WriteLine(istelienTextLogo .. ' ' .. replaceCharacters(replaceCharacters(translate('msgGameLoad'), nonbasicCharactersForGuiPrev), nonbasicCharactersForGui) .. '.')
	if (config['GAME']['GameMode'] == 2) then
		gameTimer = Timer()
		gameCountdown = gameCountdownSave
		addCallback(gameTimer, 'TimeReached', gameTimerEvent)
		gameTimer:SetTime(1,true)
	end
	
	indexPackage()
	
	-- GUI options lock
	lockMainGui(true)
	-- GUI game controls change
	extendMainGui(true)
	if (config['GAME']['TeamsLock'] == 1) then
		wMainWindow.btnManageTeams:SetEnabled(false)
	else
		wMainWindow.btnManageTeams:SetEnabled(true)
	end
	-- wMainWindow.btnStartGame:SetText(translate('txtPauseGame'))
	wMainWindow.qsStartGame:SetEnabled(false)
	wMainWindow.qsStartGame:SetVisible(false)
	wMainWindow.btnStartGame:SetEnabled(false)
	wMainWindow.btnStartGame:SetVisible(false)
	wMainWindow.qsPauseGame:SetEnabled(false)
	wMainWindow.qsPauseGame:SetVisible(false)
	wMainWindow.btnPauseGame:SetEnabled(false)
	wMainWindow.btnPauseGame:SetVisible(false)
	wMainWindow.qsResumeGame:SetEnabled(false)
	wMainWindow.qsResumeGame:SetVisible(false)
	wMainWindow.btnResumeGame:SetEnabled(false)
	wMainWindow.btnResumeGame:SetVisible(false)
	remapShortcut(wMainWindow.qsResumeFromSave, resetStartGameMessage('Saved'))
	wMainWindow.qsResumeFromSave:SetEnabled(true)
	wMainWindow.qsResumeFromSave:SetVisible(true)
	wMainWindow.btnResumeFromSave:SetEnabled(true)
	wMainWindow.btnResumeFromSave:SetVisible(true)
	wMainWindow.btnEndGame:SetEnabled(true)
	wMainWindow.btnLoadSave:SetEnabled(true)
	wMainWindow.btnLoadSave:SetText(translate('txtSaveGame'))
	wMainWindow.btnAnnounceScores:SetEnabled(true)
	wMainWindow.btnAnnounceScores:SetVisible(true)
	wMainWindow.btnPointGift:SetEnabled(true)
	wMainWindow.btnPointGift:SetVisible(true)
	if (config['GAME']['QuestionMode'] == 1) then
		wMainWindow.btnMasterScorer:SetVisible(false)
	else
		wMainWindow.btnMasterScorer:SetVisible(true)
	end
	remapShortcut(wMainWindow.qsEndQuestion, '/dance')
	wMainWindow.qsEndQuestion:SetVisible(true)
	wMainWindow.btnEndQuestion:SetVisible(true)
	wMainWindow.btnEndQuestion:SetEnabled(false)
	wMainWindow.qsAskQuestion:SetEnabled(false)
	wMainWindow.qsAskQuestion:SetVisible(false)
	wMainWindow.btnAskQuestion:SetEnabled(false)
	wMainWindow.btnAskQuestion:SetVisible(false)
	
	-- Continue rolling
	gameState = 2
	questionState = 0
	finishGame = false
	preQuestion()
end

function resumeGame()
	Turbine.Shell.WriteLine(istelienTextLogo .. ' ' .. replaceCharacters(replaceCharacters(translate('msgGameResume'), nonbasicCharactersForGuiPrev), nonbasicCharactersForGui) .. '.')
	gameState = 1	
	if (config['GAME']['GameMode'] == 2) then
		gameCountdown = gameCountdownSave
	end
	preQuestion()
	wMainWindow.qsResumeGame:SetEnabled(false)
	wMainWindow.qsResumeGame:SetVisible(false)
	wMainWindow.btnResumeGame:SetEnabled(false)
	wMainWindow.btnResumeGame:SetVisible(false)
	wMainWindow.qsResumeFromSave:SetEnabled(false)
	wMainWindow.qsResumeFromSave:SetVisible(false)
	wMainWindow.btnResumeFromSave:SetEnabled(false)
	wMainWindow.btnResumeFromSave:SetVisible(false)
	wMainWindow.qsPauseGame:SetEnabled(true)
	wMainWindow.qsPauseGame:SetVisible(true)
	wMainWindow.btnPauseGame:SetEnabled(true)
	wMainWindow.btnPauseGame:SetVisible(true)
	wMainWindow.btnLoadSave:SetEnabled(false)
end

function endGame()
	Turbine.Shell.WriteLine(istelienTextLogo .. ' ' .. replaceCharacters(replaceCharacters(translate('msgGameEnd'), nonbasicCharactersForGuiPrev), nonbasicCharactersForGui) .. '.')
	if (config['GAME']['LogGame'] == 1) then
		dataManagement(gameLog, 'Save', 'Istelien_gamelog')
	end	
	gameState = 0
	-- GUI options unlock
	lockMainGui(false)
	-- GUI game controls change
	-- wMainWindow.btnStartGame:SetText(translate('txtStartGame'))
	--remapShortcut(wMainWindow.qsStartGame, resetStartGameMessage())
	wMainWindow.btnLoadSave:SetText(translate('txtLoadGame'))
	wMainWindow.qsStartGame:SetEnabled(true)
	wMainWindow.qsStartGame:SetVisible(true)
	wMainWindow.btnStartGame:SetEnabled(true)
	wMainWindow.btnStartGame:SetVisible(true)
	wMainWindow.qsPauseGame:SetEnabled(false)
	wMainWindow.qsPauseGame:SetVisible(false)
	wMainWindow.btnPauseGame:SetEnabled(false)
	wMainWindow.btnPauseGame:SetVisible(false)
	wMainWindow.qsResumeGame:SetEnabled(false)
	wMainWindow.qsResumeGame:SetVisible(false)
	wMainWindow.btnResumeGame:SetEnabled(false)
	wMainWindow.btnResumeGame:SetVisible(false)
	wMainWindow.qsResumeFromSave:SetEnabled(false)
	wMainWindow.qsResumeFromSave:SetVisible(false)
	wMainWindow.btnResumeFromSave:SetEnabled(false)
	wMainWindow.btnResumeFromSave:SetVisible(false)
	wMainWindow.btnLoadSave:SetEnabled(true)
	wMainWindow.btnLoadSave:SetText(translate('txtLoadGame'))
	-- Many may be redundant, but it does not hurt.
	wMainWindow.btnEndGame:SetEnabled(false)
	wMainWindow.btnPointGift:SetEnabled(false)
	wMainWindow.btnPointGift:SetVisible(false)
	wMainWindow.btnMasterScorer:SetVisible(false)
	wMainWindow.qsAckCommand:SetEnabled(false)
	wMainWindow.qsAckCommand:SetVisible(false)
	wMainWindow.btnAckCommand:SetEnabled(false)
	wMainWindow.btnAckCommand:SetVisible(false)
	wMainWindow.qsEndQuestion:SetEnabled(false)
	wMainWindow.qsEndQuestion:SetVisible(false)
	wMainWindow.btnEndQuestion:SetEnabled(false)
	wMainWindow.btnEndQuestion:SetVisible(false)
	wMainWindow.qsAskQuestion:SetEnabled(false)
	wMainWindow.qsAskQuestion:SetVisible(false)
	wMainWindow.btnAskQuestion:SetEnabled(false)
	wMainWindow.btnAskQuestion:SetVisible(false)
	wMainWindow.qsWrapQuestion:SetEnabled(false)
	wMainWindow.qsWrapQuestion:SetVisible(false)
	wMainWindow.btnWrapQuestion:SetEnabled(false)
	wMainWindow.btnWrapQuestion:SetVisible(false)
	extendMainGui(false)
	shareResults(true) -- Open window with Results and button to publish results
	activateWindow(wShareResults, true)
	cleanEventMonitors()
end

-- Decide what to do after one question has ben wrapped.
function preQuestion()
	questionState = 0
	wMainWindow.btnMasterScorer:SetEnabled(false)
	local dateTable = Turbine.Engine.GetDate()
	gameLog['Date'] = addZeros(dateTable['Year'], 4) .. '-' .. addZeros(dateTable['Month'], 2) .. '-' .. addZeros(dateTable['Day'], 2)
	gameLog['Time'] = addZeros(dateTable['Hour'], 2) .. ':' .. addZeros(dateTable['Minute'], 2) .. ':' .. addZeros(dateTable['Second'], 2)
	if (config['GAME']['LogGame'] == 1) then
		dataManagement(gameLog, 'Save', 'Istelien_gamelog')
	end
	if (config['GAME']['GameMode'] == 2) then
		-- Turbine.Shell.WriteLine('Game Running. Countdown:' .. gameCountdown) -- DEBUG
	end
	if (gameState == 1) then -- Game is running normally
		-- Check if there's room for another question the game must end
		finishGame = checkGameContinue(config['GAME']['GameMode'], config['GAME']['GameTarget'])
		if (finishGame == true) then
			endGame()
		else	
			questionCounter = questionCounter + 1
			gameLog['QuestionCounter'] = questionCounter
			-- Select question and prepare answer
			selectedQuestionIndex = math.random(1, maxQuestionIndex)
			questionAuthor = quizPackage['Content'][selectedQuestionIndex]['Author']
			questionCategory = quizPackage['Content'][selectedQuestionIndex]['Category']
			questionQuestion = quizPackage['Content'][selectedQuestionIndex]['Question']
			questionAnswer = quizPackage['Content'][selectedQuestionIndex]['Answer']
			-- questionAnswer = 'Grn respuesta' -- DEBUG
			questionScore = calcQuestionScore(questionAnswer)
			local questionMessage = ''
			if (packageName ~= nil) and (packageName ~= '') then
				questionMessage = questionMessage .. packageName .. ' ' 
			end
			if (packageCategory ~= nil) and (packageCategory ~= '') then
				questionMessage = questionMessage .. '- ' .. packageCategory .. ' ' 
			end
			if (packageAuthor ~= nil) and (packageAuthor ~= '') then
				if (questionMessage == '') then
					questionMessage = replaceCharacters(translate('msgQuiz'), nonbasicCharactersInChat)
				end
				questionMessage = questionMessage .. replaceCharacters(translate('msgBy'), nonbasicCharactersInChat) .. ' ' .. packageAuthor .. '. '
			end
			questionMessage = questionMessage .. chatColours['Orange'] .. replaceCharacters(translate('msgQuestion'), nonbasicCharactersInChat) .. ' ' .. chatColours['Closure'] .. '#' .. questionCounter
			if (questionCategory ~= nil) and (questionCategory ~= '') then
				questionMessage = questionMessage .. ' - ' .. chatColours['Orange'] .. replaceCharacters(translate('msgCategory'), nonbasicCharactersInChat) .. ': ' .. chatColours['Closure'] .. questionCategory
			end
			if (questionAuthor ~= nil) and (questionAuthor ~= '') then
				questionMessage = questionMessage .. '. ' .. chatColours['Orange'] .. replaceCharacters(translate('msgAuthor'), nonbasicCharactersInChat) .. ': ' .. chatColours['Closure'] .. questionAuthor
			end
			local questionMessage =  questionMessage .. '.\n' .. questionQuestion
			-- Turbine.Shell.WriteLine(questionMessage) -- DEBUG
			-- Turbine.Shell.WriteLine('Answer:' .. questionAnswer) -- DEBUG
			
			remapShortcut(wMainWindow.qsAskQuestion, sendToChat(questionMessage, config['GAME']['GameChannel'], '', true)) -- Set chat command for firing the next question
			wMainWindow.qsAskQuestion:SetEnabled(true)
			wMainWindow.qsAskQuestion:SetVisible(true)
			wMainWindow.btnAskQuestion:SetEnabled(true)
			wMainWindow.btnAskQuestion:SetVisible(true)
		end
	else	-- Some button gas been pushed overriding game State.
		if (gameState == 0) then -- End game
			endGame()
		else
			if (gameState == 2) then -- Pause game
				if (config['GAME']['GameMode'] == 2) then
					gameCountdownSave = gameCountdown
					-- Turbine.Shell.WriteLine('Game Paused. Countdown:' .. gameCountdown) -- DEBUG
				end
				wMainWindow.qsPauseGame:SetEnabled(false)
				wMainWindow.qsPauseGame:SetVisible(false)
				wMainWindow.qsPauseGame:SetVisible(false)
				wMainWindow.qsResumeGame:SetEnabled(true)
				wMainWindow.qsResumeGame:SetVisible(true)
				wMainWindow.btnResumeGame:SetEnabled(true)
				wMainWindow.btnResumeGame:SetVisible(true)
				wMainWindow.qsAskQuestion:SetEnabled(false)
				wMainWindow.qsAskQuestion:SetVisible(false)
				wMainWindow.btnAskQuestion:SetEnabled(false)
				wMainWindow.btnAskQuestion:SetVisible(false)
			else
				-- This shouldn't happen
			end
		end
	end
end

-- Launch new question
function askQuestion()
	-- Gui changes
	wMainWindow.qsPauseGame:SetEnabled(true)
	wMainWindow.btnPauseGame:SetEnabled(true)
	wMainWindow.qsProvideHint:SetEnabled(false)
	wMainWindow.qsProvideHint:SetVisible(false)
	wMainWindow.btnProvideHint:SetEnabled(false)
	wMainWindow.btnProvideHint:SetVisible(false)
	wMainWindow.qsEndQuestion:SetEnabled(true)
	wMainWindow.btnEndQuestion:SetEnabled(true)
	remapShortcut(wMainWindow.qsEndQuestion, sendToChat(chatColours['Orange'] .. replaceCharacters(translate('msgTimesUp'), nonbasicCharactersInChat) .. chatColours['Closure'] .. '. ' .. replaceCharacters(translate('msgManualWrap'), nonbasicCharactersInChat) .. '.', config['GAME']['GameChannel'], ''))
	wMainWindow.qsAskQuestion:SetEnabled(false)
	wMainWindow.qsAskQuestion:SetVisible(false)
	wMainWindow.btnAskQuestion:SetEnabled(false)
	wMainWindow.btnAskQuestion:SetVisible(false)
	endQuestionMessage = ''

	-- Populate this question's parameters
	gameLog[questionCounter] = {}
	local dateTable = Turbine.Engine.GetDate()
	gameLog[questionCounter]['Date'] = addZeros(dateTable['Year'], 4) .. '-' .. addZeros(dateTable['Month'], 2) .. '-' .. addZeros(dateTable['Day'], 2)
	gameLog[questionCounter]['Time'] = addZeros(dateTable['Hour'], 2) .. ':' .. addZeros(dateTable['Minute'], 2) .. ':' .. addZeros(dateTable['Second'], 2)
	gameLog[questionCounter]['Question'] = questionQuestion
	gameLog[questionCounter]['Answer'] = questionAnswer
	gameLog[questionCounter]['Score'] = questionScore
	gameLog[questionCounter]['Hits'] = 0
	gameLog[questionCounter]['Scorers'] = {['Players'] = {}, ['Teams'] = {}}
	-- Create question timer
	questionTimer = Timer()
	questionCountdown = config['GAME']['QuestionTime']
	addCallback(questionTimer, 'TimeReached', questionTimerEvent)
	questionTimer:SetTime(1,true)
	questionState = 1 
end

-- Operations to be done when wrapping a question up
function endQuestion()
	questionState = 2
	-- Disable buttons
	wMainWindow.qsEndQuestion:SetVisible(false)
	wMainWindow.qsEndQuestion:SetEnabled(false)
	wMainWindow.btnEndQuestion:SetEnabled(false)
	remapShortcut(wMainWindow.qsEndQuestion, '/dance')
	wMainWindow.qsProvideHint:SetEnabled(false)
	wMainWindow.qsProvideHint:SetVisible(false)
	wMainWindow.btnProvideHint:SetEnabled(false)
	wMainWindow.btnProvideHint:SetVisible(false)
	-- End timers if they're ongoing
	questionTimer, questionTimerEvent, questionCountdown = closeTimer(questionTimer, questionTimerEvent, questionCountdown)
	extraTimer, extraTimerEvent, extraCountdown = closeTimer(extraTimer, extraTimerEvent, extraCountdown)
	
	questionScoresPrinted = false	
	if (endQuestionMessage == '') then -- No winner, so we reveal the answer
		endQuestionMessage = sendToChat(chatColours['Orange'] .. replaceCharacters(translate('msgNobodyKnew'), nonbasicCharactersInChat) .. chatColours['Closure'] .. '. ' .. replaceCharacters(translate('msgAnswerWas'), nonbasicCharactersInChat) .. ': "'.. chatColours['Orange'] .. questionAnswer .. chatColours['Closure'] ..'".', config['GAME']['GameChannel'], sender) -- String to announce it in chat
	end
	if (config['GAME']['QuestionMode'] == 1) then -- Automatic mode
		updateQuestionScoring(questionCounter, true) -- Arrange scores won in this question (if any)
		local appendScores = printQuestionScoresAll('', questionCounter, 'Append')
		if (appendScores ~= nil) then
			endQuestionMessage = endQuestionMessage .. '\n' .. appendScores
		end
		wMainWindow.qsWrapQuestion:SetEnabled(true) -- Allow advancing to next step.
		wMainWindow.qsWrapQuestion:SetVisible(true) -- Allow advancing to next step.
		wMainWindow.btnWrapQuestion:SetEnabled(true) -- Allow advancing to next step.
		wMainWindow.btnWrapQuestion:SetVisible(true) -- Allow advancing to next step.
		-- Turbine.Shell.WriteLine('EndQuestionAuto ' .. endQuestionMessage) -- DEBUG
		remapShortcut(wMainWindow.qsWrapQuestion, endQuestionMessage) -- Set up quickslot to publish the results of this question.
	else -- Manual mode
		suggestedScores = updateQuestionScoring(questionCounter, false)
		wMainWindow.btnMasterScorer:SetEnabled(true) -- Enabled button, just in case we close the window by mistake
		--[[ Get the players and scores in alphabetical order: -- DEBUG
		for playerName, scoreSuggestion in pairsByKeys(suggestedScores) do
			Turbine.Shell.WriteLine('Player: '.. playerName .. ' - SugScore: ' .. scoreSuggestion)
		end
		--]]		
		masterScorer(true) -- Launch the Master Scorer window. This will set the points and publish the results upon closing.
		activateWindow(wMasterScorer, true)
		-- The GUI user task just called should send question outcome to chat and end up firing preQuestion()
	end
end

-- Check if game must finish
function checkGameContinue(gameMode, gameTarget)

	if (gameMode ~= nil) then
		gameMode = config['GAME']['GameMode']
	end
	if (gameTarget ~= nil) then
		gameTarget = config['GAME']['GameTarget']
	end

	local gameCount
	if (gameMode == 1) then -- Game by number of questions
		gameCount = questionCounter
	else
		if (gameMode == 2) then -- Game by time (minutes)
			gameTarget = gameTarget * 60
			gameCount = gameTarget - gameCountdown
		else
			if (gameMode == 3) then -- Game by player points target
				leader,gameCount = getLeader(listAbsoluteScores('Players'))
			else -- Game by team points target
				leader,gameCount = getLeader(listAbsoluteScores('Teams'))
			end
		end
	end
	if (gameCount >= gameTarget) then
		return true
	else
		return false
	end
end

-- Start of the answer processing system
function userAnswer(inputString, sender)
	updatePlayerData(sender) -- Will add the player if it's not registered yet
	answerMatch = compareStrings(inputString, questionAnswer) -- Compare the player's answer with the valid answer
	-- debugLine = sender .. ' responde: "' .. inputString .. '"' .. '. Acert?: ' .. tostring(answerMatch) -- DEBUG
	-- Turbine.Shell.WriteLine(debugLine) -- DEBUG
	if (answerMatch == true) then
		addQuestionScorer(sender, questionCounter) -- Logs the player correct answer
		endQuestionMessage = sendToChat(chatColours['Orange'] .. sender .. chatColours['Closure'] .. ' ' .. replaceCharacters(translate('msgGotRightAnswer'), nonbasicCharactersInChat) .. ': "'.. chatColours['Orange'] .. questionAnswer .. chatColours['Closure'] ..'".', config['GAME']['GameChannel'], sender) -- String to announce it in chat
		if (config['SCORING']['MultipleHits'] == 1) then
			-- Turbine.Shell.WriteLine('Multiple Hits allowed? ' .. config['SCORING']['MultipleHits']) -- DEBUG
			-- Close question timer if it's still active
			questionTimer, questionTimerEvent, questionCountdown = closeTimer(questionTimer, questionTimerEvent, questionCountdown)
			-- Turbine.Shell.WriteLine('Going to check if extra timer is present') -- DEBUG
			if (extraCountdown == nil) then
				extraCountdown = math.floor((#questionAnswer / config['SCORING']['MultipleHitsTypeSpeed']) + 0.5) -- Round the number of seconds we'll allow an extra answer
				if (extraCountdown > 0) then
					-- Turbine.Shell.WriteLine('Multiple Hits = ' .. config['SCORING']['MultipleHits'] .. 'x (#questionAnswer = ' .. #questionAnswer .. ' / MultipleHitsTypeSpeed = ' .. config['SCORING']['MultipleHitsTypeSpeed'] .. ')  : Resultado = ' .. extraCountdown) -- DEBUG
					extraTimer = Timer()
					addCallback(extraTimer, 'TimeReached', extraTimerEvent)
					extraTimer:SetTime(1,true)
				else
					endQuestion()
				end
			else
				-- Turbine.Shell.WriteLine('Extra countdown was already running') -- DEBUG
			end
		else
			endQuestion()
		end
	end
end