#cs ----------------------------------------------------------------------------

 AutoIt Version: 3.3.14.5
 Author:         Alejandro Meirinhos 
 
 This script uses PREG-based CSV file parser by Ed Fletcher, 2007 (https://www.autoitscript.com/forum/topic/50680-simple-quick-csv-parser/), modified by FichteFoll, 2012 (https://www.autoitscript.com/forum/topic/50680-simple-quick-csv-parser/?do=findComment&comment=961109). 

 Version: 1.0 - 2024-01-05

 Script Function:
	Converts a CSV file into a LOTRO Plugin saved data file, checking if it has
	the expected field names for Istelien L quizzes.
 
 Notice:
	Copyright (C) 2023-2024 Alejandro Meirinhos from Spain.

	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with This program.  If not, see <https://www.gnu.org/licenses/>.
--]]

#ce ----------------------------------------------------------------------------

#include <Array.au3>
#include <File.au3>
#include <GUIConstantsEx.au3>
#include 'LibCsv2_mod.au3'

$bImportFile = True ; Switch for performing the final task of the program

$sImportFile = FileOpenDialog('Select a CSV file to convert', @SCRIPTDIR, 'Comma-separated values (*.csv)|Comma-separated values (*.txt)|See al files (*.*)', BitOR(1, 2))
If @error <> 0 Then
	$bImportFile = False
	Msgbox(16, 'Error','Error selecting the file to import.', 30)
Else	
	$aImportFile = _CSVReadFile($sImportFile)
	If $aImportFile = 0 Then
		$iErrorCode = @error
		Local $aCSVError[4] = [3,'File not found/openable.','File read error.','CSV format error.']
		Msgbox(16, 'File error',$aCSVError[$iErrorCode], 30)
		$bImportFile = False
	EndIf
EndIf


If $bImportFile = True Then
	_ArrayDisplay($aImportFile, 'Preview of the file to import')

	Local $aExpectedFieldNames[5][4] = [['Expected Field','Compulsory','Found','Column'],['Author','No','',''],['Category','No','',''],['Question','Yes','',''],['Answer','Yes','','']]
	; Get headers
	Local $aFieldNames[1]
	For $j = 0 To UBound($aImportFile, 2) - 1
		_ArrayAdd($aFieldNames, $aImportFile[0][$j])
	Next
	$iNumFields = _ArrayDelete($aFieldNames, 0)

	$bHeaderError = False
	For $i = 1 To UBound($aExpectedFieldNames, 1) - 1
		$indexFound = _ArraySearch($aFieldNames, $aExpectedFieldNames[$i][0], 0, 0, 1, 0, 0, 0)
		If $indexFound = -1 Then
			$aExpectedFieldNames[$i][2] = 'No'
			If $aExpectedFieldNames[$i][1] = 'Yes' Then
				$bHeaderError = True
			EndIf
		Else
			$aExpectedFieldNames[$i][2] = 'Yes'
			$aExpectedFieldNames[$i][3] = $indexFound
		EndIf
	Next
	
	If $bHeaderError = True Then
		Msgbox(48, 'File error','The first line of the file you are trying to import indicates one of more missing required fields.' & @CRLF & 'Important: field names are case-sensitive.' & @CRLF & 'The following table shows the observed status.', 30)
		_ArrayDisplay($aExpectedFieldNames, 'Expected Fields', '1:|2', 64, Default, 'Expected Field|Compulsory|Found')	
		If Msgbox(20, 'Header error', 'A quiz file with this flaw will not work.' & @CRLF & 'Do you want to continue the conversion using the wrong field names anyway?', 30) = 7 Then
			$bImportFile = false
		EndIf
	EndIf
EndIf

If $bImportFile = True Then
	$aPackageInfo = inputInfo()
	; Create file from the imported data
	Local $aExportFile[14] = ['return','{', @TAB & '[1] =', @TAB & '{', @TAB & @TAB & '["Info"] =', @TAB & @TAB & '{', @TAB & @TAB & @TAB & '["PackageName"] = "' & $aPackageInfo[1] & '",', @TAB & @TAB & @TAB & '["PackageCategory"] = "' & $aPackageInfo[2] & '",', @TAB & @TAB & @TAB & '["PackageAuthor"] = "' & $aPackageInfo[3] & '",', @TAB & @TAB & @TAB & '["PackageVersion"] = "' & $aPackageInfo[4] & '",', @TAB & @TAB & @TAB & '["PackageDate"] = "' & $aPackageInfo[5] & '"', @TAB & @TAB & '},', @TAB & @TAB & '["Content"] =', @TAB & @TAB & '{']
	$iLineCounter = UBound($aExportFile)
	For $i = 1 To Ubound($aImportFile) - 1
		$iLineCounter = _ArrayAdd($aExportFile, @TAB & @TAB & @TAB & '[' & $i & '] =')
		$iLineCounter = _ArrayAdd($aExportFile, @TAB & @TAB & @TAB & '{')
		For $j = 0 To $iNumFields - 1
			$iLineCounter = _ArrayAdd($aExportFile, @TAB & @TAB & @TAB & @TAB & '["' & $aFieldNames[$j] & '"] = "' & StringReplace(StringReplace(StringReplace($aImportFile[$i][$j], '\', '\\'), '"', '\"'), '''','\''') & '",', 0, '|', @CRLF, 1)
		Next
		$iLineCounter = _ArrayAdd($aExportFile, @TAB & @TAB & @TAB & '},')
	Next
	$iLineCounter = _ArrayAdd($aExportFile, @TAB & @TAB & '}')
	$iLineCounter = _ArrayAdd($aExportFile, @TAB & '},')
	$iLineCounter = _ArrayAdd($aExportFile, '}')
	;_ArrayDisplay($aExportFile)
	$sExportFile = FileSaveDialog('Save the converted file', @SCRIPTDIR, 'LOTRO Plugin Datafaile (*.plugindata)', BitOR(2, 16), 'MyQuiz.plugindata')
	If @error <> 0 Then
		Msgbox(16, 'Error','Error setting destination file.', 30)
	Else
		$hExportFile = FileOpen($sExportFile, 514) ; This forces to encode it in ANSI
		$bReturn = 1
		$bReturn = $bReturn * _FileWriteFromArray($hExportFile, $aExportFile, Default, Default, @CRLF)
		$bReturn = $bReturn * FileClose($hExportFile)
		#cs ; Alternative to the previous four lines to encode it in UTF-8 by default. Currently I think the file needs to be encoded in ANSI for LOTRO to properly read non-English characters without black magic and make the user's life difficult
		$bReturn = _FileWriteFromArray($sExportFile, $aExportFile, Default, Default, @CRLF)
		#ce
		If $bReturn = 1 Then
			Msgbox(64, 'Success','The quiz file has been created.', 30)
		Else
			Msgbox(16, 'Failure','There was an error exporting the quiz file.', 30)			
		EndIf
	EndIf
EndIf

Func inputInfo()
	$iGuiFontSize = 10
	$iGuiSpacing = $iGuiFontSize
	$iGuiControlFontHeight = Ceiling($iGuiFontSize * 2.35)
	$iGuiHeight = $iGuiSpacing + $iGuiControlFontHeight * 2 + $iGuiSpacing + $iGuiControlFontHeight * 2 + $iGuiSpacing + $iGuiControlFontHeight * 2 + $iGuiSpacing + $iGuiControlFontHeight * 2 + $iGuiSpacing + $iGuiControlFontHeight + $iGuiSpacing
	$iGuiWidth = $iGuiControlFontHeight * 15
	$iGuiControlWidth = $iGuiWidth - ($iGuiSpacing * 2)
	$iGuiButtonWidth = ($iGuiControlFontHeight * 5)
	$hGui = GUICreate('Package Information', $iGuiWidth, $iGuiHeight)
	
	$idLabel_PackageName = GUICtrlCreateLabel('Please enter a title or a general topic for this package: ', $iGuiSpacing, $iGuiSpacing, $iGuiControlWidth, $iGuiControlFontHeight)
		GUICtrlSetFont($idLabel_PackageName, $iGuiFontSize)
    $idInput_PackageName = GUICtrlCreateInput('', $iGuiSpacing, $iGuiSpacing + $iGuiControlFontHeight, $iGuiControlWidth, $iGuiControlFontHeight)
		GUICtrlSetFont($idInput_PackageName, $iGuiFontSize)
	
	$idLabel_PackageCategory = GUICtrlCreateLabel('Enter an optional subtitle or category for this package: ', $iGuiSpacing, $iGuiSpacing + $iGuiControlFontHeight * 2 + $iGuiSpacing, $iGuiControlWidth, $iGuiControlFontHeight)
		GUICtrlSetFont($idLabel_PackageCategory, $iGuiFontSize)
    $idInput_PackageCategory = GUICtrlCreateInput('', $iGuiSpacing, $iGuiSpacing + $iGuiControlFontHeight * 2 + $iGuiSpacing + $iGuiControlFontHeight, $iGuiControlWidth, $iGuiControlFontHeight)	
		GUICtrlSetFont($idInput_PackageCategory, $iGuiFontSize)
	
	$idLabel_PackageAuthor = GUICtrlCreateLabel('Please introduce the author of this package: ', $iGuiSpacing, $iGuiSpacing + $iGuiControlFontHeight * 4 + $iGuiSpacing + $iGuiSpacing, $iGuiControlWidth, $iGuiControlFontHeight)
		GUICtrlSetFont($idLabel_PackageAuthor, $iGuiFontSize)
    $idInput_PackageAuthor = GUICtrlCreateInput('', $iGuiSpacing, $iGuiSpacing + $iGuiControlFontHeight * 4 + $iGuiSpacing + $iGuiSpacing + $iGuiControlFontHeight, $iGuiControlWidth, $iGuiControlFontHeight)	
		GUICtrlSetFont($idInput_PackageAuthor, $iGuiFontSize)
	
	$idLabel_PackageVersion = GUICtrlCreateLabel('Package Version: ', $iGuiSpacing, $iGuiSpacing + $iGuiControlFontHeight * 6 + $iGuiSpacing * 3, $iGuiWidth / 2 - $iGuiSpacing * 2, $iGuiControlFontHeight)
		GUICtrlSetFont($idLabel_PackageVersion, $iGuiFontSize)
    $idInput_PackageVersion = GUICtrlCreateInput('', $iGuiSpacing, $iGuiSpacing + $iGuiControlFontHeight * 7 + $iGuiSpacing * 3, $iGuiWidth / 2 - $iGuiSpacing * 2, $iGuiControlFontHeight)	
		GUICtrlSetFont($idInput_PackageVersion, $iGuiFontSize)
		GUICtrlSetData($idInput_PackageVersion, '1.0')
	
	$idLabel_PackageDate = GUICtrlCreateLabel('Package Date: ', $iGuiWidth / 2, $iGuiSpacing + $iGuiControlFontHeight * 6 + $iGuiSpacing * 3, $iGuiWidth / 2 - $iGuiSpacing, $iGuiControlFontHeight)
		GUICtrlSetFont($idLabel_PackageDate, $iGuiFontSize)
    $idInput_PackageDate = GUICtrlCreateInput('', $iGuiWidth / 2, $iGuiSpacing + $iGuiControlFontHeight * 7 + $iGuiSpacing * 3, $iGuiWidth / 2 - $iGuiSpacing, $iGuiControlFontHeight)	
		GUICtrlSetFont($idInput_PackageDate, $iGuiFontSize)
		GUICtrlSetData($idInput_PackageDate, @YEAR & '-' & @MON & '-' & @MDAY)
	
	$idButton_Continue = GUICtrlCreateButton('Continue', ($iGuiWidth / 2) - $iGuiButtonWidth - ($iGuiSpacing / 2), $iGuiSpacing + $iGuiControlFontHeight * 8 + $iGuiSpacing * 4, $iGuiButtonWidth, $iGuiControlFontHeight)
	$idButton_Cancel = GUICtrlCreateButton('Cancel', ($iGuiWidth / 2) + ($iGuiSpacing / 2), $iGuiSpacing + $iGuiControlFontHeight * 8 + $iGuiSpacing * 4, $iGuiButtonWidth, $iGuiControlFontHeight)
	
	GUISetState(@SW_SHOW, $hGui)
	$bExitGui = False
	Local $aUserInput[1] = [0]
	While $bExitGui = False
		$idEvent = GUIGetMsg(1)
		Select
			Case $idEvent[0] = $GUI_EVENT_CLOSE
				Exit
			Case $idEvent[0] = $idButton_Continue
				$aUserInput[0] = _ArrayAdd($aUserInput, GUICtrlRead($idInput_PackageName))
				$aUserInput[0] = _ArrayAdd($aUserInput, GUICtrlRead($idInput_PackageCategory))
				$aUserInput[0] = _ArrayAdd($aUserInput, GUICtrlRead($idInput_PackageAuthor))
				$aUserInput[0] = _ArrayAdd($aUserInput, GUICtrlRead($idInput_PackageVersion))
				$aUserInput[0] = _ArrayAdd($aUserInput, GUICtrlRead($idInput_PackageDate))
				$bExitGui = True				
			Case $idEvent[0] = $idButton_Cancel
				Exit
		EndSelect
	Wend
	
	GUIDelete($hGui)
	
	Return $aUserInput

EndFunc