--[[
    Istelien L (implementation of Istelien quiz engine in LUA with the aim of using it in LOTRO)
	Copyright (C) 2023-2024 Alejandro Meirinhos from Spain.

	This file is part of Istelien L.

	Istelien L is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Istelien L is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Istelien L.  If not, see <https://www.gnu.org/licenses/>.
--]]


-- Assign nominal score to a question
function calcQuestionScore(inputString)

	scoreValue = 1 -- Default (ScoreMode = 1)
	wordList,numWords = stringSplit(inputString)
	if (config['SCORING']['ScoreMode'] == 2) then -- One point per word
			scoreValue = numWords
	else
		if (config['SCORING']['ScoreMode'] == 3) then -- Custom	points per number of words
			if (config['SCORING']['CustomScores'][numWords] ~= nil) then
				scoreValue = config['SCORING']['CustomScores'][numWords]
			else
				scoreValue = config['SCORING']['CustomScores']['More']
			end
		end
	end

	return scoreValue
end

-- Logs that a player answered right to the specific question
function addQuestionScorer(playerName, questionNumber) 
	questionNumber = tonumber(questionNumber)
	if (gameLog[questionNumber]['Scorers']['Players'][playerName] == nil) then -- Only add if it's the player's first correct answer

		gameLog[questionNumber]['Hits'] = gameLog[questionNumber]['Hits'] + 1
		gameLog[questionNumber]['Scorers']['Players'][playerName] = 
		{
			['Order'] = gameLog[questionNumber]['Hits'],
			['Points'] = 0 -- Dummy value. We'll update it later
		}
		if (gameParticipants['Players'][playerName]['Team']) ~= nil then -- Add points to team if applicable
			local playerTeam = gameParticipants['Players'][playerName]['Team']
			if (gameLog[questionNumber]['Scorers']['Teams'][playerTeam] == nil) then
				gameLog[questionNumber]['Scorers']['Teams'][playerTeam] = 
				{
					['Order'] = gameLog[questionNumber]['Hits'],
					['Points'] = 0 -- Dummy value. We'll update it later
				}
			end
		end
	end
end

-- Calculates and assigns points for all scoring players for the specific question
function updateQuestionScoring(questionNumber, update)
	if (update == nil) then
		update = true
	end
	questionNumber = tonumber(questionNumber)
	local scorersCount = gameLog[questionNumber]['Hits'] -- Get number of winners
	local scorersList = {}
	if scorersCount > 0 then 
		local answerScore = gameLog[questionNumber]['Score'] -- Get nominal value of the question
		for playerName,details in pairs(gameLog[questionNumber]['Scorers']['Players']) do
			local playerPoints
			if (gameLog[questionNumber]['Scorers']['Players'][playerName]['Order'] == 1) then -- The first winner takes all points
				playerPoints = answerScore
			else -- Different scoring options
				if (config['SCORING']['MultipleHitsDivisor'] == 2) then
					playerPoints = round(answerScore / gameLog[questionNumber]['Scorers']['Players'][playerName]['Order'], 2) -- Each new winner has a decreasing score
				else
					if (config['SCORING']['MultipleHitsDivisor'] == 1) then
						playerPoints = round(answerScore / scorersCount, 2)-- All secondary winners get the same score, but the more winners are, the lesser the score
					else
						playerPoints = round(answerScore / config['SCORING']['MultipleHitsDivisor'], 2) -- Al secondary winners get the same score, but it's fixed by the game configuration
					end
				end
			end
			-- Turbine.Shell.WriteLine('Player points from updateQuestionScoring: ' .. tostring(playerPoints)) -- DEBUG
			scorersList[playerName] = playerPoints
			if (update == true) then
				addScore(playerName, playerPoints, questionNumber) -- Add player points to the general participants table
			end
		end
	end

	return scorersList
end

-- Add a score obtained by a player to its individual and team record. If questionNumber is provided it logs it to the gameLog too.
function addScore(playerName, points, questionNumber)
	if (questionNumber == nil) then
		questionNumber = 0 -- Point gift
	end
	local dateTable = Turbine.Engine.GetDate()
	points = tonumber(points)
	if (gameParticipants['Players'][playerName] == nil) then -- Player wasn't previously registered.
		updatePlayerData(playerName) -- Add player without team
		gameParticipants['Players'][playerName]['Points'] = points -- Add its first points
		if (questionNumber ~= 0) then -- Add player points to game log.
			gameLog[questionNumber]['Scorers']['Players'][playerName]['Points'] = points
		else -- Point gift
			pointGiftCounter = pointGiftCounter + 1
			gameLog[questionNumber][pointGiftCounter] = {['Date'] = addZeros(dateTable['Year'], 4) .. '-' .. addZeros(dateTable['Month'], 2) .. '-' .. addZeros(dateTable['Day'], 2), ['Time'] = addZeros(dateTable['Hour'], 2) .. ':' .. addZeros(dateTable['Minute'], 2) .. ':' .. addZeros(dateTable['Second'], 2), ['Players'] = {[playerName] = {['Points'] = points}}}
			-- As this is a separate action, we'll update the general date and time of the log too
			gameLog['Date'] = addZeros(dateTable['Year'], 4) .. '-' .. addZeros(dateTable['Month'], 2) .. '-' .. addZeros(dateTable['Day'], 2)
			gameLog['Time'] = addZeros(dateTable['Hour'], 2) .. ':' .. addZeros(dateTable['Minute'], 2) .. ':' .. addZeros(dateTable['Second'], 2)
		end
	else
		gameParticipants['Players'][playerName]['Points'] = gameParticipants['Players'][playerName]['Points'] + points -- Increment point quantity
		if (questionNumber ~= 0) then
			gameLog[questionNumber]['Scorers']['Players'][playerName]['Points'] = points -- Add player points to game log.
			if (gameParticipants['Players'][playerName]['Team']) ~= nil then -- Add points to team if applicable
				local playerTeam = gameParticipants['Players'][playerName]['Team'] 
				if (gameParticipants['Teams'][playerTeam] ~= nil) then
					if (gameParticipants['Teams'][playerTeam]['Points'] == nil) then  -- First points for this team
						gameParticipants['Teams'][playerTeam]['Points'] = points
					else
						gameParticipants['Teams'][playerTeam]['Points'] = gameParticipants['Teams'][playerTeam]['Points'] + points
					end
					if (questionNumber ~= nil) then -- Add team points to game log.
						if (gameLog[questionNumber]['Scorers']['Teams'][playerTeam]['Points'] == nil) then
							gameLog[questionNumber]['Scorers']['Teams'][playerTeam]['Points'] = points
						else
							gameLog[questionNumber]['Scorers']['Teams'][playerTeam]['Points'] = gameLog[questionNumber]['Scorers']['Teams'][playerTeam]['Points'] + points
						end	
					end
				else
					Turbine.Shell.WriteLine('ERROR: Player ' .. playerName .. ' is registered as member of a nonexisting team: ' .. playerTeam .. '.') -- DEBUG. This shouldn't happen
				end
			end			
		else -- point gift
			pointGiftCounter = pointGiftCounter + 1
			gameLog[0]['PointGiftCounter'] = pointGiftCounter
			gameLog[questionNumber][pointGiftCounter] = {['Date'] = addZeros(dateTable['Year'], 4) .. '-' .. addZeros(dateTable['Month'], 2) .. '-' .. addZeros(dateTable['Day'], 2), ['Time'] = addZeros(dateTable['Hour'], 2) .. ':' .. addZeros(dateTable['Minute'], 2) .. ':' .. addZeros(dateTable['Second'], 2), ['Players'] = {[playerName] = {['Points'] = points}}}
			gameLog['Date'] = addZeros(dateTable['Year'], 4) .. '-' .. addZeros(dateTable['Month'], 2) .. '-' .. addZeros(dateTable['Day'], 2)
			gameLog['Time'] = addZeros(dateTable['Hour'], 2) .. ':' .. addZeros(dateTable['Minute'], 2) .. ':' .. addZeros(dateTable['Second'], 2)
			if (gameParticipants['Players'][playerName]['Team']) ~= nil then -- Add points to team if applicable
				local playerTeam = gameParticipants['Players'][playerName]['Team']
				gameLog[questionNumber][pointGiftCounter]['Teams'] = {[playerTeam] = {['Points'] = points}}
				-- Add points to the participant table
				if (gameParticipants['Teams'][playerTeam]['Points'] == nil) then  -- First points for this team
					gameParticipants['Teams'][playerTeam]['Points'] = points
				else
					gameParticipants['Teams'][playerTeam]['Points'] = gameParticipants['Teams'][playerTeam]['Points'] + points
				end
			end
			-- As this is a separate action, we'll update the general date and time of the log too
			gameLog['Date'] = addZeros(dateTable['Year'], 4) .. '-' .. addZeros(dateTable['Month'], 2) .. '-' .. addZeros(dateTable['Day'], 2)
			gameLog['Time'] = addZeros(dateTable['Hour'], 2) .. ':' .. addZeros(dateTable['Minute'], 2) .. ':' .. addZeros(dateTable['Second'], 2)
		end
	end
end

-- Add manual score to a team. As this is a very specific case I preferred to have it separated from the general addScore() function to not complicate it even more
function addTeamScore(teamName, points)
	local dateTable = Turbine.Engine.GetDate()
	pointGiftCounter = pointGiftCounter + 1
	gameLog[0]['PointGiftCounter'] = pointGiftCounter
	gameLog[0][pointGiftCounter] = {['Date'] = addZeros(dateTable['Year'], 4) .. '-' .. addZeros(dateTable['Month'], 2) .. '-' .. addZeros(dateTable['Day'], 2), ['Time'] = addZeros(dateTable['Hour'], 2) .. ':' .. addZeros(dateTable['Minute'], 2) .. ':' .. addZeros(dateTable['Second'], 2), ['Teams'] = {[teamName] = {['Points'] = points}}}
	gameLog['Date'] = addZeros(dateTable['Year'], 4) .. '-' .. addZeros(dateTable['Month'], 2) .. '-' .. addZeros(dateTable['Day'], 2)
	gameLog['Time'] = addZeros(dateTable['Hour'], 2) .. ':' .. addZeros(dateTable['Minute'], 2) .. ':' .. addZeros(dateTable['Second'], 2)
	-- Add points to the participant table
	if (gameParticipants['Teams'][teamName]['Points'] == nil) then  -- First points for this team
		gameParticipants['Teams'][teamName]['Points'] = points
	else
		gameParticipants['Teams'][teamName]['Points'] = gameParticipants['Teams'][teamName]['Points'] + points
	end
end

-- Set all participant scores to 0 for a fresh start
function resetScores()
	for k,v in pairs(gameParticipants['Players']) do
		gameParticipants['Players'][k]['Points'] = 0
	end
	for k,v in pairs(gameParticipants['Teams']) do
		gameParticipants['Teams'][k]['Points'] = 0
	end
end

-- See who has the most points, and how many
function getLeader(scorersList)
	local maxPoints = 0
	local leader = ''
	if (scorersList ~= nil) then
		for i,v in pairs(scorersList) do
			if v > maxPoints then
				maxPoints = v
				leader = i
			end
		end
		-- Turbine.Shell.WriteLine('Lder: ' .. leader .. ' - Puntos: ' .. maxPoints) -- DEBUG
	end
	return leader, maxPoints
end

-- Retrieves a table containing ParticipantName = Score
function listAbsoluteScores(participantsType)
	if (participantsType == nil) then
		participantsType = 'Players'
	end
	local participantScores = {}
	local counter = 0
	for k,v in pairs(gameParticipants[participantsType]) do
		if (gameParticipants[participantsType][k]['Points'] ~= 0) then
			counter = counter + 1
			participantScores[k] = gameParticipants[participantsType][k]['Points']
		end
	end
	if (counter == 0) then
		participantScores = nil
	end
	return participantScores
end

-- Provides the list of participants sorted by scores
function getStandings(participantsType)
	local scoreList = listAbsoluteScores(participantsType)
	local standings = {}
	if (scoreList ~= nil) then
		standings = sortedArrayFromTable(scoreList)
	else
		standings = nil
	end
	return standings
end

-- Retrieves a table containing ParticipantName = Score for a specific question
function listQuestionScores(questionNumber, participantsType)
	if (participantsType == nil) then
		participantsType = 'Players'
	end
	questionNumber = tonumber(questionNumber)
	local questionScorers = gameLog[questionNumber]['Scorers'][participantsType]
	local participantScores = {}
	local countParticipants = 0
	for i,v in pairs(questionScorers) do
		countParticipants = countParticipants + 1
		participantScores[i] = v['Points']
	end
	if (countParticipants) == 0 then
		participantScores = nil
	end
	return participantScores
end